<'
# Copyright (c) 2012-2014,2017-2020,2023 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Wed Mar 14 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]

lang c
'>
#include "autoconf/acheader.h"

#include <stdio.h>

#include "h2devLib.h"
#include "posterLib.h"

#include "<"$comp">/c/client.h"
#include "<"$comp">_msglib.h"
#include "<"$comp">_portlib.h"
#include "private.h"
#include "serialize.h"


/* --- local data ---------------------------------------------------------- */

<'foreach s [$component services] {'>
int	genom_<"$comp">_<"[$s name]">_encode(char *buffer, int size,
		char *dst, int maxsize);
int	genom_<"$comp">_<"[$s name]">_decode(char *buffer, int size,
		char *dst, int maxsize);
<'}'>
<'foreach p [$component ports out] {'>
int	genom_<"$comp">_<"[$p name]">_decode(char *buffer,
		<"[[$p datatype] argument reference dst]">);
<'}'>


/* <"[--- ${comp}_abort -------------------------------------------------]"> */

static void
genom_<"$comp">_client_abort_cb(genom_client h, int rqstid, int done,
                       genom_event report, void *output, const void *exdetail,
                       void *cb_data)
{
  (void)done; (void)report; (void)output;
  (void)exdetail; (void)cb_data; /* fix -Wunused-parameter */

  (void)pocolibs_clean(h, rqstid);
}

genom_event
genom_<"$comp">_client_abort(genom_client h, int rqstid)
{
  int abrtid;

  if (rqstid < 0 || rqstid >= h->nactive)
    return genom_no_such_activity(&h->context);
  if (h->active_rqst[rqstid].status == FREE)
    return genom_no_such_activity(&h->context);
  if (h->active_rqst[rqstid].status == WAITING_INTERMED_REPLY)
    return genom_syserr(&(genom_syserr_detail){.code = EAGAIN}, &h->context);

  return pocolibs_sendrqst(
    h, <"$COMP">_abort_activity_RQSTID,
    genom_<"$comp">_client_abort_activity_info,
    &h->active_rqst[rqstid].aid, genom_<"$comp">_abort_activity_encode,
    FALSE,
    (void (*)(void *))genom_<"$comp">_client_abort_activity_init_output,
    genom_<"$comp">_abort_activity_decode,
    sizeof(struct genom_<"$comp">_abort_activity_output),
    NULL, genom_<"$comp">_client_abort_cb, NULL,
    &abrtid);
}

<'foreach s [$component services] {'>

/* <"[--- ${comp}_[$s name]_rqst ----------------------------------------]"> */

genom_event
genom_<"$comp">_client_<"[$s name]">_rqst(
  genom_client h,
  const struct genom_<"$comp">_<"[$s name]">_input *in,
  <"$comp">_<"[$s name]">_cb sentcb, <"$comp">_<"[$s name]">_cb donecb,
  void *cb_data,
  int *rqstid)
{
<'  set noack [catch {$s task}]'>
  assert(h);

  return pocolibs_sendrqst(
    h, <"$COMP">_<"[$s name]">_RQSTID,
    genom_<"$comp">_client_<"[$s name]">_info,
    in, genom_<"$comp">_<"[$s name]">_encode,
    <"[expr ${noack} ? "FALSE" : "TRUE"]">,
    (void (*)(void *))genom_<"$comp">_client_<"[$s name]">_init_output,
    genom_<"$comp">_<"[$s name]">_decode,
    sizeof(struct genom_<"$comp">_<"[$s name]">_output),
    (genom_request_cb)sentcb, (genom_request_cb)donecb, cb_data,
    rqstid);
}


/* <"[--- ${comp}_[$s name]_result --------------------------------------]"> */

genom_event
genom_<"$comp">_client_<"[$s name]">_result(
  genom_client h, int rqstid, genom_event *report,
  struct genom_<"$comp">_<"[$s name]">_output **out, void **exdetail)
{
  struct genom_client_request *a;
  genom_event s;
  void *p;

  if (rqstid < 0 || rqstid >= h->nactive)
    return genom_syserr(&(genom_syserr_detail){.code = ENOENT}, &h->context);

  a = &h->active_rqst[rqstid];
  if (a->info != genom_<"$comp">_client_<"[$s name]">_info)
    return genom_syserr(&(genom_syserr_detail){.code = ENOMSG}, &h->context);

  s = pocolibs_result(h, rqstid, report, &p, exdetail);
  if (s) return s;
  if (out) *out = p;
  return genom_ok;
}

<'}'>

<'foreach p [$component ports out] {'>

/* <"[--- ${comp}_[$p name]_port ----------------------------------------]"> */

genom_event
genom_<"$comp">_client_<"[$p name]">_port(
  genom_client h,
<'if {"multiple" in [$p kind]} {'>
  const char *path,
<'}'>
  <"[[$p datatype] argument reference data]">)
{
  POSTER_ID *id;
  int searched = 0;
  char *addr;
  int s;

  /* cached id */
<'  switch -- [$p kind] {'>
<'    simple {'>
  id = &h->ports-><"[$p name]">.id;
<'    }'>
<'    multiple {'>
  unsigned int i;

  for(i = 0; i < h->ports-><"[$p name]">.n; i++) {
    if (!strcmp(h->ports-><"[$p name]">.h[i].name, path)) break;
    if (h->ports-><"[$p name]">.h[i].name[0] == 0) {
      /* reuse empty slot */
      snprintf(h->ports-><"[$p name]">.h[i].name,
               sizeof(h->ports-><"[$p name]">.h[i].name), "%s", path);
      break;
    }
  }
  if (i >= h->ports-><"[$p name]">.n) {
    void *a = realloc(
      h->ports-><"[$p name]">.h, (i+1) * sizeof(*h->ports-><"[$p name]">.h));
    if (!a)
      return genom_syserr(
        &(genom_syserr_detail){.code = errno}, &h->context);

    h->ports-><"[$p name]">.h = a;
    snprintf(h->ports-><"[$p name]">.h[i].name,
             sizeof(h->ports-><"[$p name]">.h[i].name), "%s", path);
    h->ports-><"[$p name]">.h[i].id = NULL;
    h->ports-><"[$p name]">.n = i+1;
  }

  id = &h->ports-><"[$p name]">.h[i].id;
<'    }'>
<'  }'>

  do {
    if (!*id) {
      char name[H2_DEV_MAX_NAME];
      int n;
<'  switch -- [$p kind] {'>
<'    simple {'>
      n = snprintf(name, sizeof(name), "%s/<"[$p name]">", h->instance);
<'    }'>
<'    multiple {'>
      n = snprintf(
        name, sizeof(name), "%s/<"[$p name]">/%s", h->instance, path);
<'    }'>
<'  }'>
      if (n <= 0 || n >= (signed)sizeof(name)) {
<'  if {"multiple" in [$p kind]} {'>
        h->ports-><"[$p name]">.h[i].name[0] = 0;
        h->ports-><"[$p name]">.h[i].id = NULL;
<'  }'>
        return genom_syserr(
          &(genom_syserr_detail){.code = ENAMETOOLONG}, &h->context);
      }

      if (posterFind(name, id) != OK) {
        genom_mwerr_detail d;
        char msg[sizeof(d.what)];
        h2getErrMsg(errnoGet(), msg, sizeof(msg));
        if (snprintf(d.what, sizeof(d.what), "%s port: %s", name, msg) < 0)
          strncpy(d.what, "internal error", sizeof(d.what));
<'  if {"multiple" in [$p kind]} {'>
        h->ports-><"[$p name]">.h[i].name[0] = 0;
        h->ports-><"[$p name]">.h[i].id = NULL;
<'  }'>
        return genom_mwerr(&d, &h->context);
      }
      searched = 1;
    }

    if (posterTake(*id, POSTER_READ) != OK) {
      genom_mwerr_detail d;
      char msg[sizeof(d.what)];
      int code = errnoGet();
      if (!searched && code == S_posterLib_POSTER_CLOSED) {
        *id = NULL; continue /* try to refetch poster id */;
      }

      h2getErrMsg(code, msg, sizeof(msg));
<'  switch -- [$p kind] {'>
<'    simple {'>
      if (snprintf(d.what, sizeof(d.what),
                   "%s/<"[$p name]"> port: %s", h->instance, msg) < 0)
        strncpy(d.what, "internal error", sizeof(d.what));
<'    }'>
<'    multiple {'>
      if (snprintf(d.what, sizeof(d.what),
                   "%s/%s port: %s", h->instance, path, msg) < 0)
        strncpy(d.what, "internal error", sizeof(d.what));
<'    }'>
<'  }'>
      return genom_mwerr(&d, &h->context);
    }

    break;
  } while(1);

  addr = posterAddr(*id);
  s = genom_<"$comp">_<"[$p name]">_decode(addr, data);
  posterGive(*id);

  if (s)
    return genom_syserr(&(genom_syserr_detail){.code = s}, &h->context);
  return genom_ok;
}

<'}'>
